class
	IDENTIFIER_TRANSFORMER

inherit
	TRANSFORMATION_STEP
	AST_DEFAULT_VISITOR
		redefine
			end_visit_translation_unit,
			end_visit_function_declaration,
			end_visit_struct_declaration,
			end_visit_enum_declaration
		end

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		do
			transformer.tree.accept (Current)
			Result := True
		end

	description: STRING
		do
			Result := "Transforming names of fields, variables and functions"
		end

feature {ANY} -- AST_DEFAULT_VISITOR

	end_visit_translation_unit (node: AST_TRANSLATION_UNIT)
		local
			feature_names: HASH_SET [STRING]
			project_struct_names: HASH_SET [STRING]
			system_struct_names: HASH_SET [STRING]
			project_enum_names: HASH_SET [STRING]
			system_enum_names: HASH_SET [STRING]
		do
			create feature_names.make (16)
			create project_struct_names.make (16)
			create system_struct_names.make (16)
			create project_enum_names.make (16)
			create system_enum_names.make (16)

			across node.declarations as c loop
				if attached {AST_FUNCTION_DECLARATION} c.item as tmp then
					adapt_identifier (tmp.identifier, feature_names)
				elseif attached {AST_VARIABLE_DECLARATION} c.item as tmp then
					adapt_identifier (tmp.identifier, feature_names)
				elseif attached {AST_STRUCT_DECLARATION} c.item as tmp then
					if tmp.identifier.starts_with ("__anonstruct_") or tmp.identifier.starts_with ("__anonunion_") then
						tmp.identifier.make_from_string (tmp.identifier.substring (
								tmp.identifier.index_of ('_', 3) + 1,
								tmp.identifier.last_index_of ('_', tmp.identifier.count) -1
							))
						if tmp.identifier.is_empty then -- e.g. from "__anonstruct_28" (wget)
							tmp.identifier.make_from_string ("ANONYMOUS")
						end
					end
					tmp.identifier.to_upper
					if tmp.source_info.is_project_source then
						adapt_identifier (tmp.identifier, project_struct_names)
					else
						adapt_identifier (tmp.identifier, system_struct_names)
					end
				elseif attached {AST_ENUM_DECLARATION} c.item as tmp then
					if tmp.identifier.starts_with ("__anonenum_") then
						tmp.identifier.make_from_string (tmp.identifier.substring (
								tmp.identifier.index_of ('_', 3) + 1,
								tmp.identifier.last_index_of ('_', tmp.identifier.count) -1
							))
						if tmp.identifier.is_empty then -- e.g. from "__anonstruct_28" (wget)
							tmp.identifier.make_from_string ("ANONYMOUS")
						end
					end
					tmp.identifier.to_upper
					if tmp.source_info.is_project_source then
						adapt_identifier (tmp.identifier, project_enum_names)
					else
						adapt_identifier (tmp.identifier, system_enum_names)
					end
				end
			end
		end

	end_visit_function_declaration (node: AST_FUNCTION_DECLARATION)
		local
			existing_identifiers: HASH_SET [STRING]
		do
			create existing_identifiers.make (node.parameter_list.parameters.count)
			across node.parameter_list.parameters as c loop
				if attached {AST_VARIABLE_DECLARATION} c.item as tmp then
					tmp.identifier.prepend ("a_")
					adapt_identifier (tmp.identifier, existing_identifiers)
				end
			end

			if node.locals /= Void then
				create existing_identifiers.make (node.locals.count)
				across node.locals as c loop
					c.item.identifier.prepend ("l_")
					adapt_identifier (c.item.identifier, existing_identifiers)
				end
			end
		end

	end_visit_struct_declaration (node: AST_STRUCT_DECLARATION)
		local
			existing_identifiers: HASH_SET [STRING]
		do
			create existing_identifiers.make (node.fields.count)
			across node.fields as c loop
				if attached c.item.identifier as identifier then
					adapt_identifier (identifier, existing_identifiers)
				end
			end
		end

	end_visit_enum_declaration (node: AST_ENUM_DECLARATION)
		local
			existing_identifiers: HASH_SET [STRING]
		do
			create existing_identifiers.make (node.fields.count)
			across node.fields as c loop
				adapt_identifier (c.item.identifier, existing_identifiers)
			end
		end

feature {NONE}

	adapt_identifier (identifier: STRING; existing_identifiers: HASH_SET [STRING])
			-- Changes a C identifier to an Eiffel identifier.
		require
			identifier /= Void and then not identifier.is_empty
			existing_identifiers /= Void
		local
			i: INTEGER_32
			lc: STRING
		do
			-- strip leading underscores
			from
				i := 1
			until
				identifier[i] /= '_'
			loop
				i := i + 1
			end
			if i > 1 then
				identifier.remove_head (i - 1)
			end

			-- handle duplication and keywords
			from
				lc := identifier.as_lower
				i := 1
			until
				not existing_identifiers.has (lc) and
				not eiffel_keywords.has (lc)
			loop
				i := i + 1
				lc := identifier.as_lower + i.out
			end
			if i > 1 then
				identifier.append_integer (i)
			end

			-- don't allow identifiers starting with `c_'
			if identifier.as_lower.starts_with ("c_") then
				identifier.remove_head (1)
				identifier.prepend ("c2")
			end

			existing_identifiers.extend (identifier.as_lower)
		end

	eiffel_keywords: HASH_SET [STRING]
		once ("PROCESS")
			create Result.make (100)
			Result.extend ("agent")
			Result.extend ("alias")
			Result.extend ("all")
			Result.extend ("and")
			Result.extend ("as")
			Result.extend ("assign")
			Result.extend ("attribute")
			Result.extend ("bit")
			Result.extend ("check")
			Result.extend ("class")
			Result.extend ("convert")
			Result.extend ("copy")
			Result.extend ("create")
			Result.extend ("current")
			Result.extend ("debug")
			Result.extend ("deferred")
			Result.extend ("dispose")
			Result.extend ("do")
			Result.extend ("do_nothing")
			Result.extend ("else")
			Result.extend ("elseif")
			Result.extend ("end")
			Result.extend ("ensure")
			Result.extend ("equal")
			Result.extend ("expanded")
			Result.extend ("export")
			Result.extend ("external")
			Result.extend ("false")
			Result.extend ("feature")
			Result.extend ("from")
			Result.extend ("frozen")
			Result.extend ("if")
			Result.extend ("implies")
			Result.extend ("inherit")
			Result.extend ("inspect")
			Result.extend ("invariant")
			Result.extend ("like")
			Result.extend ("local")
			Result.extend ("loop")
			Result.extend ("not")
			Result.extend ("note")
			Result.extend ("obsolete")
			Result.extend ("old")
			Result.extend ("once")
			Result.extend ("only")
			Result.extend ("or")
			Result.extend ("out")
			Result.extend ("precursor")
			Result.extend ("prefix")
			Result.extend ("redefine")
			Result.extend ("rename")
			Result.extend ("require")
			Result.extend ("rescue")
			Result.extend ("result")
			Result.extend ("retry")
			Result.extend ("select")
			Result.extend ("separate")
			Result.extend ("some")
			Result.extend ("then")
			Result.extend ("true")
			Result.extend ("tuple")
			Result.extend ("undefine")
			Result.extend ("unique")
			Result.extend ("until")
			Result.extend ("variant")
			Result.extend ("void")
			Result.extend ("when")
			Result.extend ("xor")

			-- reserved for helpers
			Result.extend ("l_aaf")
			Result.extend ("l_pos")
		end

end
