class
	JUMP_STATEMENT_TRANSFORMER

inherit
	TRANSFORMATION_STEP
	AST_DEFAULT_VISITOR
		redefine
			visit_function_declaration,
			end_visit_function_declaration,
			end_visit_labeled_statement,
			end_visit_jump_statement,
			end_visit_eiffel_jump
		end
	NAME_UTILITIES

feature {ANY} -- TRANSFORMATION_STEP

	process_impl: BOOLEAN
		do
			transformer.tree.accept (Current)
			Result := True
		end

	description: STRING
		do
			Result := "Transforming jump statements"
		end

feature {NONE}

	goto_jump_states: HASH_TABLE [AST_EIFFEL_JUMP_STATE, STRING]

feature {ANY} -- AST_DEFAULT_VISITOR

	visit_function_declaration (node: AST_FUNCTION_DECLARATION): BOOLEAN
		do
			create goto_jump_states.make (16)
			Result := True
		end

	end_visit_function_declaration (node: AST_FUNCTION_DECLARATION)
		local
			l_agent_visitor: AST_AGENT_VISITOR
			l_end_of_function_jump: AST_EIFFEL_JUMP
			l_ti: TUPLE [NATURAL_32]
		do
			-- After visiting the whole function, we know if there are jump statements or not.
			-- If yes, add "returns". If no, delete return statements; not needed.
			create l_agent_visitor
			l_agent_visitor.end_visit_jump_statement_agent := agent (ag_node: AST_JUMP_STATEMENT; ag_has_jump_statements: BOOLEAN)
				do
					if ag_has_jump_statements then
						replace_node (ag_node, create {AST_EIFFEL_JUMP}.make (Void, "return"))
					else
						delete_node (ag_node)
					end
				end (?, node.has_jump_statements)
			node.accept (l_agent_visitor)

			if node.has_jump_statements then
				insert_adjacent_statement (node.statements.statements.first, create {AST_EIFFEL_JUMP_STATE}.make_comment (0, "execution start:"), False)

				if node.return_statements_removed then
					-- no more return statements -> add a "return" at the end of the function
					create l_end_of_function_jump.make (Void, "end of function")
					insert_adjacent_statement (node.statements.statements.last, l_end_of_function_jump, True)
					insert_jump_state (create {AST_EIFFEL_JUMP_STATE}.make (0), l_end_of_function_jump, False)
				end

				-- split statements:
				create l_agent_visitor
				l_agent_visitor.end_visit_if_statement_agent := agent (ag_node: AST_IF_STATEMENT)
					do
						if ag_node.requires_split then
							split_if_statement (ag_node)
						end
					end
				l_agent_visitor.end_visit_while_statement_agent := agent (ag_node: AST_WHILE_STATEMENT)
					do
						-- process from inside out, otherwise no longer existing nodes will be processed
						if ag_node.requires_split then
							split_iteration_statement (ag_node)
						end
					end
				node.accept (l_agent_visitor)

				-- number AST_EIFFEL_JUMP_STATEs:
				create l_ti
				create l_agent_visitor
				l_agent_visitor.end_visit_eiffel_jump_state_agent := agent (ag_node: AST_EIFFEL_JUMP_STATE; ag_ti: TUPLE [i: NATURAL_32])
					local
						l_succeeding_jump_state: AST_EIFFEL_JUMP_STATE
					do
						if ag_node.parent /= Void then -- ignore already deleted duplicates
							ag_node.index.set_item (ag_ti.i)
							from
								l_succeeding_jump_state ?= get_adjacent_statement (ag_node, True)
							until
								l_succeeding_jump_state = Void
							loop
								if l_succeeding_jump_state.comment /= Void then
									ag_node.append_comment (l_succeeding_jump_state.comment)
								end
								l_succeeding_jump_state.index.set_item (ag_ti.i) -- update referencing AST_EIFFEL_JUMPs
								delete_node (l_succeeding_jump_state)
								l_succeeding_jump_state ?= get_adjacent_statement (ag_node, True)
							end
							ag_ti.i := ag_ti.i + 1
						end
					end (?, l_ti)
				node.accept (l_agent_visitor)
			end

			goto_jump_states := Void
		end

	end_visit_labeled_statement (node: AST_LABELED_STATEMENT)
		do
			if not goto_jump_states.has (node.identifier) then
				goto_jump_states[node.identifier] := create {AST_EIFFEL_JUMP_STATE}.make_comment (0, node.identifier + ":")
			end
			insert_adjacent_statement (node, node.statement.ast_twin, True)
			insert_jump_state (goto_jump_states[node.identifier], node, True)
			delete_node (node)
		end

	end_visit_jump_statement (node: AST_JUMP_STATEMENT)
		do
			inspect
				node.type
			when C_kw_goto then
				handle_goto (node)
			when C_kw_break then
				handle_break (node)
			when C_kw_continue then
				handle_continue (node)
			when C_kw_return then
				handle_return (node)
			end
		end

	end_visit_eiffel_jump (node: AST_EIFFEL_JUMP)
		do
			-- jump from `SWITCH_TRANSFORMER'
			stop_execution_after (node)
		end

feature {NONE}

	handle_goto (node: AST_JUMP_STATEMENT)
		require
			node /= Void and then node.type = C_kw_goto
		do
			stop_execution_after (node)
			if not goto_jump_states.has (node.goto_identifier) then
				goto_jump_states[node.goto_identifier] := create {AST_EIFFEL_JUMP_STATE}.make_comment (0, node.goto_identifier + ":")
			end
			replace_node (node, create {AST_EIFFEL_JUMP}.make (goto_jump_states[node.goto_identifier], "goto " + node.goto_identifier))
		end

	handle_break (node: AST_JUMP_STATEMENT)
		require
			node /= Void and then node.type = C_kw_break
		local
			cur: AST_NODE
			loop_statement: AST_WHILE_STATEMENT
			jump_state: AST_EIFFEL_JUMP_STATE
		do
			stop_execution_after (node)
			-- get the loop we `break' -> loop_statement
			from
				cur := node.parent
			until
				loop_statement /= Void
			loop
				if attached {AST_WHILE_STATEMENT} cur as tmp then
					loop_statement := tmp
				end
				cur := cur.parent
			end
			-- create the eiffel jump
			if attached {AST_EIFFEL_JUMP_STATE} get_adjacent_statement (loop_statement, True) as tmp then
				jump_state := tmp
			else
				create jump_state.make_comment (0, "loop done")
				insert_jump_state (jump_state, loop_statement, True)
			end
			replace_node (node, create {AST_EIFFEL_JUMP}.make (jump_state, "break"))
		end

	handle_continue (node: AST_JUMP_STATEMENT)
		require
			node /= Void and then node.type = C_kw_continue
		local
			cur: AST_NODE
			loop_statement: AST_WHILE_STATEMENT
			jump_state: AST_EIFFEL_JUMP_STATE
		do
			stop_execution_after (node)
			-- get the loop we `continue' -> loop_statement
			from
				cur := node.parent
			until
				loop_statement /= Void
			loop
				if attached {AST_WHILE_STATEMENT} cur as tmp then
					loop_statement := tmp
				end
				cur := cur.parent
			end
			-- create the eiffel jump
			if attached {AST_EIFFEL_JUMP_STATE} loop_statement.statements.statements.last as tmp then
				jump_state := tmp
			else
				create jump_state.make_comment (0, "loop body end")
				insert_jump_state (jump_state, loop_statement.statements.statements.last, True)
			end
			replace_node (node, create {AST_EIFFEL_JUMP}.make (jump_state, "continue"))
		end

	handle_return (node: AST_JUMP_STATEMENT)
		require
			node /= Void and then node.type = C_kw_return
		do
			stop_execution_after (node)
			if node.return_expression /= Void then
				insert_adjacent_statement (node, create {AST_EXPRESSION_STATEMENT}.make (create {AST_ASSIGNMENT_EXPRESSION}.make (
						create {AST_EIFFEL_EXPRESSION}.make ("Result", node.return_expression.type_binding),
						node.return_expression.ast_twin)), False)
			end
		end

	insert_jump_state (jump_state: AST_EIFFEL_JUMP_STATE; anchor: AST_STATEMENT; after: BOOLEAN)
		require
			jump_state /= Void
			anchor /= Void
		local
			cur: AST_NODE
		do
			insert_adjacent_statement (anchor, jump_state, after)
			from
				cur := jump_state.parent
			until
				cur = Void
			loop
				if attached {AST_IF_STATEMENT} cur as tmp then
					tmp.requires_split := True
				elseif attached {AST_WHILE_STATEMENT} cur as tmp then
					tmp.requires_split := True
				elseif attached {AST_FUNCTION_DECLARATION} cur as tmp then
					tmp.has_jump_statements := True
					cur := Void
				end
				if cur /= Void then
					cur := cur.parent
				end
			end
		end

	stop_execution_after (node: AST_STATEMENT)
		require
			node /= Void
		local
			cur: AST_EXPRESSION_STATEMENT
		do
			if attached {AST_COMPOUND_STATEMENT} node.parent as compound_statement then
				if compound_statement.statements.last /= node or attached {AST_WHILE_STATEMENT} compound_statement.parent then
					if not attached {AST_EIFFEL_JUMP_STATE} get_adjacent_statement (node, True) then
						insert_jump_state (create {AST_EIFFEL_JUMP_STATE}.make (0), node, True)
					end
				elseif attached {AST_IF_STATEMENT} compound_statement.parent as if_statement then
					stop_execution_after (if_statement)
				elseif attached {AST_FUNCTION_DECLARATION} compound_statement.parent then
					-- execution will stop automatically
				else
					log ("[stop_execution_after] Unexpected case?!")
				end
			elseif attached {AST_LABELED_STATEMENT} node.parent as labeled_statement then
				stop_execution_after (labeled_statement)
			elseif
				attached {AST_EIFFEL_EXPRESSION} node.parent as eiffel_expression and then
				attached {AST_EXPRESSION_STATEMENT} eiffel_expression.parent as expression_statement
			then
				-- we're in a switch head (inspect); it might not end automatically (e.g. `switch(i) { foo; case 1: bar; }'
				from
					cur ?= get_adjacent_statement (expression_statement, True)
				until
					attached {AST_EIFFEL_EXPRESSION} cur.expression as ee and then ee.mixed.count = 1 and then ee.is_string (1, "end")
				loop
					cur ?= get_adjacent_statement (cur, True)
				end
				stop_execution_after (cur)
			else
				log (generator + ".stop_execution_after: Unexpected parent: " + node.parent.generator + " (" + node.parent.c_code + ")")
			end
		end

feature {NONE} -- helpers for "split statements"

	split_if_statement (node: AST_IF_STATEMENT)
		require
			node /= Void and then node.requires_split
		local
			then_jump_state: AST_EIFFEL_JUMP_STATE
			else_jump_state: AST_EIFFEL_JUMP_STATE
			after_jump_state: AST_EIFFEL_JUMP_STATE
			target_jump_state: AST_EIFFEL_JUMP_STATE -- either points to `else_jump_state' or `after_jump_state'
		do
			-- create the jump states
			create then_jump_state.make_comment (0, "if:")
			if
				attached {AST_IF_ELSE_STATEMENT} node as stmt and then
				stmt.else_statements /= Void
			then
				create else_jump_state.make_comment (0, "else:")
			end
			if attached {AST_EIFFEL_JUMP_STATE} get_adjacent_statement (node, True) as tmp then
				after_jump_state := tmp
			else
				create after_jump_state.make_comment (0, "if done")
				insert_adjacent_statement (node, after_jump_state, True)
			end
			if else_jump_state /= Void then
				target_jump_state := else_jump_state
			else
				target_jump_state := after_jump_state
			end
			then_jump_state.indent_update := 1
			after_jump_state.indent_update := -1
			-- update the code
			delete_node (node)
			insert_adjacent_statement (
					after_jump_state,
					create {AST_IF_STATEMENT}.make_if_short (
							create_negated_expression (node.condition),
							create {AST_EIFFEL_JUMP}.make (target_jump_state, "skip if block")
						),
					False
				)
			insert_adjacent_statement (after_jump_state, then_jump_state, False)
			across node.then_statements.statements as c loop
				insert_adjacent_statement (after_jump_state, c.item.ast_twin, False)
			end
			if
				else_jump_state /= Void and then
				attached {AST_IF_ELSE_STATEMENT} node as stmt
			then
				insert_adjacent_statement (
						after_jump_state,
						create {AST_EIFFEL_JUMP_STATE}.make_comment (0, "if block done, skip else block:"),
						False
					)
				insert_adjacent_statement (
						after_jump_state,
						create {AST_EIFFEL_JUMP}.make (after_jump_state, Void),
						False
					)
				insert_adjacent_statement (after_jump_state, else_jump_state, False)
				across stmt.else_statements.statements as c loop
					insert_adjacent_statement (after_jump_state, c.item.ast_twin, False)
				end
			end
		end

	split_iteration_statement (node: AST_WHILE_STATEMENT)
		require
			node /= Void and then node.requires_split
		local
			before_jump_state: AST_EIFFEL_JUMP_STATE
			body_jump_state: AST_EIFFEL_JUMP_STATE
			after_jump_state: AST_EIFFEL_JUMP_STATE
		do
			-- create the jump states
			if attached {AST_EIFFEL_JUMP_STATE} get_adjacent_statement (node, False) as tmp then
				before_jump_state := tmp
			else
				create before_jump_state.make_comment (0, "loop condition:")
				insert_adjacent_statement (node, before_jump_state, False)
			end
			create body_jump_state.make_comment (0, "loop body:")
			if attached {AST_EIFFEL_JUMP_STATE} get_adjacent_statement (node, True) as tmp then
				after_jump_state := tmp
			else
				create after_jump_state.make_comment (0, "loop done")
				insert_adjacent_statement (node, after_jump_state, True)
			end
			body_jump_state.indent_update := 1
			after_jump_state.indent_update := -1
			-- update the code
			delete_node (node)
			insert_adjacent_statement (after_jump_state, create {AST_IF_ELSE_STATEMENT}.make_if_short (
					create_negated_expression (node.condition),
					create {AST_EIFFEL_JUMP}.make (after_jump_state, "skip loop")), False)
			insert_adjacent_statement (after_jump_state, body_jump_state, False)
			across node.statements.statements as c loop
				insert_adjacent_statement (after_jump_state, c.item.ast_twin, False)
			end
			insert_adjacent_statement (after_jump_state, create {AST_EIFFEL_JUMP}.make (before_jump_state, "back to loop condition"), False)
		end

end
